pcr_relay_bridge_detect() {
    PCR_RELAY_PORT="`ubus list 'dli.serial.*'|sed -e 's/^dli.serial.//g'`"
    if [ -n "$PCR_RELAY_PORT" ]; then
        PCR_RELAY_BAUD_RATE=`stty </dev/$PCR_RELAY_PORT |awk '/^speed / {print $2;exit(0);}'`
        PCR_RELAY_BRIDGE="dli.serial.$PCR_RELAY_PORT"
    else
        PCR_RELAY_BRIDGE=""
    fi
    PCR_RELAY_BRIDGE_SUBORDINATE_PID=""
    PCR_RELAY_BRIDGE_RESTORATION_METHOD=""
}

pcr_relay_config_detect() {
    PCR_RELAY_PORT=`uci get misc.settings.relay_ports`
}

pcr_relay_detect() {
    pcr_relay_bridge_detect
    if [ -z "$PCR_RELAY_PORT" ] ; then
        pcr_relay_config_detect
    fi
    if [ -z "$PCR_RELAY_PORT" ] && [ -x /usr/bin/avr_detect ] && [ -x /bin/stty ] ; then
        pcr_relay_avr_detect
    fi
    if [ -z "$PCR_RELAY_PORT" ] ; then
        PCR_RELAY_PORT="ttyATH0"
    fi
}

PCR_RELAY_BRIDGE_SUBORDINATE_DEFAULT_NAME="dli.serial.pcr"

pcr_relay_subordinate_bridge_create() {
    if [ -z "$PCR_RELAY_BRIDGE" ]; then
        ubus_serial_bridge "$PCR_RELAY_BRIDGE_SUBORDINATE_DEFAULT_NAME" </dev/"$PCR_RELAY_PORT" > /dev/"$PCR_RELAY_PORT" &
        PCR_RELAY_BRIDGE_SUBORDINATE_PID=$!
        PCR_RELAY_BRIDGE="$PCR_RELAY_BRIDGE_SUBORDINATE_DEFAULT_NAME"
        while ! ubus call "$PCR_RELAY_BRIDGE" send '{"line":""}'; do
            sleep 1;
        done
    fi
}

pcr_relay_subordinate_bridge_destroy() {
    if [ -n "$PCR_RELAY_BRIDGE_SUBORDINATE_PID" ]; then
        kill "$PCR_RELAY_BRIDGE_SUBORDINATE_PID"
        PCR_RELAY_BRIDGE_SUBORDINATE_PID=""
        PCR_RELAY_BRIDGE=""
    fi
}

pcr_relay_unbridged_configure() {
    stty </dev/"$PCR_RELAY_PORT" raw ispeed $PCR_RELAY_BAUD_RATE ospeed $PCR_RELAY_BAUD_RATE >/dev/null
}

pcr_relay_bridged_scope() {
    if [ -n "$PCR_RELAY_BRIDGE" ]; then
       "$@"
    else
        pcr_relay_subordinate_bridge_create
        "$@"
        pcr_relay_subordinate_bridge_destroy
        pcr_relay_unbridged_configure
    fi
}

pcr_relay_unbridged_scope() {
    if [ -z "$PCR_RELAY_BRIDGE" ]; then
        "$@"
    else
        if [ -n "$PCR_RELAY_BRIDGE_SUBORDINATE_PID" ]; then
            pcr_relay_subordinate_bridge_destroy
            pcr_relay_unbridged_configure
            "$@"
            pcr_relay_subordinate_bridge_create
        elif [ -x /etc/init.d/ubus_serial_bridge ]; then
            /etc/init.d/ubus_serial_bridge stop
            sleep 1
            if ubus call "$PCR_RELAY_BRIDGE" send '{"line":""}' && [ -x /usr/bin/ubus_serial_bridge ] && grep -qs /usr/bin/ubus_serial_bridge /usr/libexec/ubus_serial_bridge_run ; then
                # NB: Doing permanent FS changes
                ln /usr/bin/ubus_serial_bridge /usr/sbin/ubus_serial_bridge
                sed -e 's@/usr/bin/ubus_serial_bridge@/usr/sbin/ubus_serial_bridge@g' </usr/libexec/ubus_serial_bridge_run > /usr/libexec/ubus_serial_bridge_run.new
                chmod +x /usr/libexec/ubus_serial_bridge_run.new
                mv -f /usr/libexec/ubus_serial_bridge_run.new /usr/libexec/ubus_serial_bridge_run
                sync
                rm /usr/bin/ubus_serial_bridge
                cat > /usr/bin/ubus_serial_bridge <<EOF
#!/bin/sh
while true; do sleep 86400;done
EOF
                chmod +x /usr/bin/ubus_serial_bridge
                killall ubus_serial_bridge
                sleep 5 # Waiting for procd to notice
                rm /usr/bin/ubus_serial_bridge
            fi
            PCR_RELAY_BRIDGE=""
            pcr_relay_unbridged_configure
            "$@"
            /etc/init.d/ubus_serial_bridge start
            PCR_RELAY_BRIDGE="dli.serial.$PCR_RELAY_PORT"
            while ! ubus call "$PCR_RELAY_BRIDGE" send '{"line":""}'; do
                sleep 1;
            done
            sleep 1
        else
            ORIGINAL_INITTAB=`mktemp`
            cp /etc/inittab "$ORIGINAL_INITTAB"
            sed -i -e "/^$PCR_RELAY_PORT:/d" /etc/inittab
            killall -HUP init
            killall ubus_serial_bridge
            mv -f "$ORIGINAL_INITTAB" /etc/inittab
            PCR_RELAY_BRIDGE=""
            pcr_relay_unbridged_configure
            "$@"
            killall -HUP init
            PCR_RELAY_BRIDGE="dli.serial.$PCR_RELAY_PORT"
            while ! ubus call "$PCR_RELAY_BRIDGE" send '{"line":""}'; do
                sleep 1;
            done
            sleep 1
        fi
    fi
}

pcr_reset_disabled_scope() {
    if [ -f /etc/hotplug.d/button/000-pcr-ignore-reset ]; then
        "$@"
    else
        cat > /etc/uci-defaults/000-pcr-no-longer-ignore-reset <<"EOF"
rm -f /etc/hotplug.d/button/000-pcr-ignore-reset
EOF
        cat > /etc/hotplug.d/button/000-pcr-ignore-reset <<"EOF"
if [ "$BUTTON" = "reset" ]; then
    BUTTON="reset-ignored"
EOF
        "$@"
        rm -f /etc/hotplug.d/button/000-pcr-ignore-reset
        rm -f /etc/uci-defaults/000-pcr-no-longer-ignore-reset
    fi
}

pcr_relay_send() {
    if [ -n "$PCR_RELAY_PORT" ]; then
        if [ -n "$PCR_RELAY_BRIDGE" ]; then
            ubus call "$PCR_RELAY_BRIDGE" send '{"line":"'"$@"'"}'
        else
            echo -n -e "$@\r\n" >"/dev/$PCR_RELAY_PORT"
        fi
    fi
}

pcr_relay_query_internal () {
    cat <"/dev/$PCR_RELAY_PORT" >/dev/null&
    PID=$!
    sleep 1
    kill $PID
    awk "/^\+$1/ { print; exit;} " <"/dev/$PCR_RELAY_PORT" &
    PID=$!
    pcr_relay_send '$'"$1$2"
    sleep 1
    kill $PID
}

pcr_relay_query () {
    pcr_relay_unbridged_scope pcr_relay_query_internal "$@"
}

pcr_relay_get_version () {
    pcr_relay_send '$U1'
    pcr_relay_query S| sed -e 's/.*ver=\([^ ]*\).*/\1/g'
}

pcr_relay_get_watchdog_cycle_count () {
    pcr_relay_send '$U1'
    pcr_relay_query W| sed -e 's/.*count=\([0-9]*\).*/\1/g'
}

pcr_reset_ui () {
    pcr_relay_send '$L1'
    pcr_relay_send '$E0'
}

pcr_relay_ensure_link () {
    TRY=10
    if [ -z "$PCR_RELAY_BRIDGE" ]; then
        pcr_relay_unbridged_configure
    fi
    if [ -z "$PCR_RELAY_ENSURE_LINK_DISABLE" ]; then
        while [ "$TRY" -gt 0 ] && [ -z "`pcr_relay_query U 1`" ]; do
            if [ "$TRY" -le 5 ] && [ -n "$PCR_RELAY_ENSURE_LINK_APPLY_FORCE" ]; then
                # NB: fragment copied from recent flash_avr_firmware.
                echo 23 > /sys/class/gpio/export
                echo out >/sys/class/gpio/gpio23/direction
                echo 0 >/sys/class/gpio/gpio23/value
                sleep 1
                echo 1 >/sys/class/gpio/gpio23/value
                echo 23 >/sys/class/gpio/unexport
            fi
            sleep 1
            TRY="$(($TRY-1))"
        done
    fi
    [ "$TRY" -gt 0 ]
}

pcr_message () {
    pcr_relay_send '$X'
    pcr_relay_send '$D'"$1"
    pcr_relay_send '$C0,1'
    pcr_relay_send '$D'"$2"
}

pcr_info_message () {
    echo '{ '"$1"'}'
    shift 1
    pcr_message "$@"
    sleep 1
}

pcr_success_message () {
    echo '{+'"$1"'}'
    shift 1
    pcr_message "$@"
    sleep 1
}

pcr_failure_message () {
    echo '{-'"$1"'}'
    shift 1
    pcr_message "$@"
    sleep 1
}

pcr_relay_enumerate_user_pids() {
    for P in /proc/*; do
        for F in $P/fd/*; do
            if [ "$F" -ef "/dev/$PCR_RELAY_PORT" ]; then
                echo "${P##/proc/}"
                break
            fi
        done
    done
}

pcr_relay_ensure_controlled_use_only() {
    if [ -z "$PCR_RELAY_BRIDGE" ]; then
        pcr_relay_enumerate_user_pids | while read PID; do
            kill "$PID"
        done
        sleep 1
        pcr_relay_enumerate_user_pids | while read PID; do
            kill -KILL "$PID"
        done
        sleep 1
        pcr_relay_enumerate_user_pids | while read PID; do
            echo "Process $PID (`xargs -0 echo </proc/$PID/cmdline`) using the relay would not die" >&2
            return 1
        done
    fi
    return 0
}
